# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logger"""
from graphrepo.singleton import Singleton


class Logger(metaclass=Singleton):
  def __init__(self, *args, **kwargs):
    """Default init"""

  def log(self, exception):
    """Logs exceptions and prints it to console
    :param exception: Exception type from Python
    """
    print('[EXCEPTION]: {}'.format(exception))

  def log_and_raise(self, exception):
    """Logs, prints and raises exception
    :param exception: Python Exception object
    """
    self.log(exception)
    raise exception



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.drillers import Driller
from graphrepo.mappers import CSVMapper
from graphrepo.miners import CommitMiner


class TestCSVMapper:
    """Most data is indexed when indexing a commmit
    so this class tests indexing for multiple models"""

    def test_csv_mapper(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        n_matcher = NodeMatcher(test_driller.graph)
        r_matcher = RelationshipMatcher(test_driller.graph)

        com_miner = CommitMiner(test_driller.graph, n_matcher, r_matcher)
        mapper = CSVMapper()

        commits = com_miner.get_all()
        mapped_commits = mapper.map(commits)
        assert mapped_commits.shape == (8, 9)

        c_files = com_miner.get_commit_files(
            'ad98f8594c15b1ebc4be4f20d849bcc0edf69ec574c33dfd84b7792d')
        c_csv = mapper.map(c_files)
        assert c_csv.shape == (3, 5)

        test_driller.clean()



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default Parent class for drillers
"""
import json
import pika

from abc import abstractmethod
from datetime import datetime
from py2neo import Graph
from pydriller import RepositoryMining

import graphrepo.utils as utl
from graphrepo.config import Config
from graphrepo.drillers.queue_driller import QueueDriller
import graphrepo.drillers.batch_utils as b_utl
from graphrepo.logger import Logger

LG = Logger()


class RabbitDriller(QueueDriller):
    """RabbitDriller class - parses a git repo and publishes
    the data in a queue every n commits
    """

    def connect_queue(self):
        """Establishes a connection to queue"""
        try:
            credentials = pika.PlainCredentials(
                self.queue['username'], self.queue['password'])
            self.con_parameters = pika.ConnectionParameters(self.queue['host'],
                                                            self.queue['port'],
                                                            self.queue['vhost'],
                                                            credentials)
            connection = pika.BlockingConnection(
                self.con_parameters)
            channel = connection.channel()

            channel.queue_declare(queue=self.queue['queue'], durable=True)
            return connection, channel
        except Exception as e:
            raise e

    def send_index_data(self, data):
        """Indexes data"""
        try:
            connection, channel = self.connect_queue()
            channel.basic_publish(
                exchange='',
                routing_key=self.queue['queue'],
                body=json.dumps(data),
                properties=pika.BasicProperties(
                    delivery_mode=2,  # make message persistent
                ))
            connection.close()
        except Exception as e:
            raise e



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" This module uses pydriller to search a repository
and indexes it in neo4j
"""
from diskcache import Cache
from datetime import datetime
from py2neo import Graph
from pydriller import RepositoryMining

import graphrepo.utils as utl
import graphrepo.drillers.batch_utils as b_utl
from graphrepo.config import Config
from graphrepo.drillers.drill_cache import DrillCacheSequential
from graphrepo.drillers.default import DefaultDriller
from graphrepo.logger import Logger

LG = Logger()


class Driller(DefaultDriller):
    """Drill class - parses a git repo and uses the models
    to index everything in Neo4j. This class is a singleton
    because it holds the connection to Neo4j in self.graph
    """

    def index_batch(self, **kwargs):
        """Indexes data extracted by drill_batch of from
        disk in Neo4j
        :param kwargs: data keys and values (see the drill_batch return)
        """
        try:
            self.config.check_config()
            self._check_connection()
            b_utl.index_all(
                self.graph, batch_size=self.config.ct.batch_size, **kwargs)
        except Exception as exc:
            LG.log_and_raise(exc)
        else:
            return

    def index_from_file(self, file_path):
        """Reads a file and indexes the data in Neo4j
        :param file_path: the path of the JSON file with data
        """
        try:
            data_ = utl.load_json(file_path)
            self.index_batch(**data_)
        except Exception as exc:
            LG.log_and_raise(exc)
        else:
            return



import os

from graphrepo.drillers.default import DefaultDriller
import graphrepo.drillers.db_init as db_init

from py2neo.database import Schema


class TestDBInit:
    def test_hash_constraints(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = DefaultDriller(os.path.join(folder, 'cnfg_simple.yml'))

        db_init.create_hash_constraints(test_driller.graph)

        schm = Schema(test_driller.graph)

        labels = ["Developer", "Branch", "Commit", "File", "Method"]

        for l in labels:
            c = schm.get_uniqueness_constraints(l)
            assert len(c) == 1

        # clean
        for l in labels:
            schm.drop_uniqueness_constraint(l, 'hash')

    def test_indices(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = DefaultDriller(os.path.join(folder, 'cnfg_simple.yml'))

        db_init.create_indices(test_driller.graph, hash_index=True)

        schm = Schema(test_driller.graph)

        index_authors = schm.get_indexes("Developer")
        assert len(index_authors) == 1

        index_branch = schm.get_indexes("Branch")
        assert len(index_branch) == 2

        index_commits = schm.get_indexes("Commit")
        assert len(index_commits) == 2

        index_files = schm.get_indexes("File")
        assert len(index_files) == 2

        index_methods = schm.get_indexes("Method")
        assert len(index_methods) == 2

        # clean
        schm.drop_index("Developer", "hash")
        schm.drop_index("Branch", "hash")
        schm.drop_index("Branch", "project_id")
        schm.drop_index("Commit", "hash")
        schm.drop_index("Commit", "project_id")
        schm.drop_index("File", "hash")
        schm.drop_index("File", "project_id")
        schm.drop_index("Method", "hash")
        schm.drop_index("Method", "project_id")



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.drillers.driller import Driller
from graphrepo.miners.method import MethodMiner


class TestMethodMiner:
    def test_get_all(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        n_matcher = NodeMatcher(test_driller.graph)
        r_matcher = RelationshipMatcher(test_driller.graph)

        m_miner = MethodMiner(test_driller.graph, n_matcher, r_matcher)

        all_methods = m_miner.get_all()
        assert len(all_methods) == 5
        m_hash = '45ce8dcd8b0cd8ed42e592ce828ab6418e7c79713b8dc99805bcb7ea'
        met = m_miner.query(hash=m_hash)
        assert met['name'] == 'get_name'

        history = m_miner.get_change_history(m_hash)
        assert len(history) == 2

        test_driller.clean()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module mines files and contains all related Neo4j queries"""

from graphrepo.miners.default import DefaultMiner


class MethodMiner(DefaultMiner):
    def __init__(self, graph, node_matcher, rel_matcher, *args, **kwargs):
        super().__init__(graph, node_matcher, rel_matcher, *args, **kwargs)

    def query(self, **kwargs):
        """Searches for a method using the arguments in kwargs.
        If no kwargs are given it returns the first method found
        """
        return self.node_matcher.match("Method", **kwargs).first()

    def get_all(self):
        """Returns all node of type Method
        :return: list of method
        """
        return self.node_matcher.match("Method")

    def get_change_history(self, method_hash):
        """Returns all UpdateMethod relationships
          :param method_hash: method unique identifier
          :param dic: optional; boolean for converting data to dictionary
          or returning it as py2neo records - the py2neo raw
          records can be used in mappers
          :return: list of UpdateMethod relationships / dics
          """
        query = """MATCH ()-[r:UpdateMethod]->(m: Method{{hash: "{0}"}})
        RETURN distinct r
          """.format(method_hash)
        dt_ = self.graph.run(query)
        return [dict(x['r']) for x in dt_.data()]



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.drillers.driller import Driller
from graphrepo.miners.file import FileMiner


class TestFileMiner:
    def test_get_all(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        n_matcher = NodeMatcher(test_driller.graph)
        r_matcher = RelationshipMatcher(test_driller.graph)

        f_miner = FileMiner(test_driller.graph, n_matcher, r_matcher)

        all_files = f_miner.get_all()
        assert len(all_files) == 6

        # get readme file
        readme = f_miner.query(name='README.MD')
        assert readme['name'] == 'README.MD'

        # get file history
        f_hash = 'f85f4af5b20ddd617f93da13c7789a65fb972e68a8d634d5f253abab'
        update_history = f_miner.get_change_history(f_hash)
        assert len(update_history) == 3

        # test file get methods
        current_m = f_miner.get_current_methods(f_hash)
        assert len(current_m) == 2

        test_driller.clean()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from graphrepo.utils import parse_config


class TestUtils:
    def test_parse_config(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        neo, project = parse_config(os.path.join(folder, 'cnfg_init.yml'))
        assert neo['db_url'] == 'localhost'
        assert neo['db_user'] == 'neo4j'
        assert project['repo'] == 'tests/gr-test'



from .commit import CommitMiner
from .default import DefaultMiner
from .developer import DeveloperMiner
from .file import FileMiner
from .mine_manager import MineManager
from .method import MethodMiner



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from graphrepo.drillers.queue_driller import QueueDriller


# class TestQueueDriller:
# def test_indexing(self):
#     folder = os.path.dirname(os.path.abspath(__file__))
#     test_driller = QueueDriller(os.path.join(folder, 'cnfg_init.yml'))
#     test_driller.drill_batch()
#     records = [r for r in test_driller.graph.run(
#         "MATCH(n) RETURN n")]
#     assert len(records) == 22

#     test_driller.clean()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import yaml
from graphrepo.miners import MineManager
from datetime import datetime


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config', default='examples/configs/pydriller.yml', type=str)
    return parser.parse_args()


def main():
    args = parse_args()

    start = datetime.now()
    miner = MineManager(config_path=args.config)

    # get all nodes and relationships from the manager
    nodes, rels = miner.get_all_data()
    print("The DB has a total of {} nodes and {} relationships".format(
        len(nodes), len(rels)))
    print("All data took: {}".format(datetime.now() - start))

    # get all commits
    commits = miner.commit_miner.get_all()
    print("The DB has a total of {} commits".format(len(commits)))

    # get all developers
    devs = miner.dev_miner.get_all()
    print("The DB has a total of {} developers".format(len(devs)))

    # get all files
    files = miner.file_miner.get_all()
    print("The DB has a total of {} files".format(len(files)))


if __name__ == '__main__':
    main()



import graphrepo.drillers.batch_utils as utils


def create_hash_constraints(graph):
    """Creates uniqueness constratins on nodes' hash"""
    query = """CREATE CONSTRAINT ON (n: {}) ASSERT n.hash IS UNIQUE"""
    nodes = ["Developer", "Branch", "Commit", "File", "Method"]
    for n in nodes:
        q = query.format(n)
        graph.run(q)


def create_indices(graph, hash_index=True):
    """Initializes all indexes for database"""
    if hash_index:
        utils.create_index_authors(graph)
    utils.create_index_branches(graph, hash_index)
    utils.create_index_commits(graph, hash_index)
    utils.create_index_files(graph, hash_index)
    utils.create_index_methods(graph, hash_index)



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


###
# This file assumes the project from the config file was already indexed
###
import argparse
import os
import pandas as pd
import plotly.express as px

from datetime import datetime
from graphrepo.miners import MineManager
from graphrepo.utils import parse_config


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', default='configs/pydriller.yml', type=str)
    return parser.parse_args()


def main():
    args = parse_args()
    mine_manager = MineManager(config_path=args.config)

    file_miner = mine_manager.file_miner
    file_ = file_miner.query(pproject_id=mine_manager.config.ct.project_id,
                             name="commit.py")
    updated_file_rels = file_miner.get_change_history(file_['hash'])

    # sort update relationships and transform data for plotting
    updated_file_rels.sort(key=lambda x: x['timestamp'])

    complexity = [x['complexity'] for x in updated_file_rels]
    nloc = [x['nloc'] for x in updated_file_rels]
    dts = [datetime.fromtimestamp(x['timestamp']) for x in updated_file_rels]

    fig = px.line(pd.DataFrame({'date': dts, 'complexity': complexity}),
                  x='date', y='complexity',
                  title='Complexity over time for the commit.py file')
    fig.show()

    fig_2 = px.line(pd.DataFrame({'date': dts, 'nloc': nloc}),
                    x='date', y='nloc', title="NLOC over time for the commit.py file")
    fig_2.show()


if __name__ == '__main__':
    main()



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default Parent class for drillers
"""
from abc import abstractmethod
from datetime import datetime
from py2neo import Graph
from pydriller import RepositoryMining

import graphrepo.utils as utl
from graphrepo.config import Config
from graphrepo.drillers.driller import Driller
import graphrepo.drillers.batch_utils as b_utl
from graphrepo.logger import Logger

LG = Logger()


class QueueDriller(Driller):
    """QueueDriller class - parses a git repo and publishes
    the data in a queue every n commits
    """

    def __init__(self, neo, project, queue):
        """Initializes the properties of this class
        :param neo:
        :param project:
        :param queue:
        """
        # TODO: validate inputs
        try:
            self.project, self.queue = project, queue
            self.config = Config()
            self.graph = None
            self.config.configure(**neo, **self.project)
            # self._connect()
        except Exception as exc:
            LG.log_and_raise(exc)

    @abstractmethod
    def connect_queue(self):
        """Establishes a connection to queue"""
        raise NotImplementedError

    @abstractmethod
    def send_index_data(self, data):
        """Indexes data"""
        raise NotImplementedError

    def drill_batch(self, index=True, save_path=None):
        """Extracts data from a software repository, with the option
        of saving it on diks and indexing it in Neo4j
        :param index: optional, if True, the data is indexed in Neo4j
        :param save_path: optional, if given, the data is stored on dik
        :returns: dictionary with all data
        """
        start = datetime.now()
        print('Driller started at: \t', start)
        commits, parents, devs, dev_com, branches,\
            branches_com, files, com_files, \
            methods, files_methods, com_methods = \
            [], [], [], [], [], [], [], [], [], [], []
        commit_index = 0
        for commit in \
            RepositoryMining(self.config.ct.repo,
                             since=self.config.ct.start_date,
                             to=self.config.ct.end_date).traverse_commits():

            self.drill_commit(commit, commits, parents, devs, dev_com, branches,
                              branches_com, files, com_files,
                              methods, files_methods, com_methods)

            if commit_index == self.queue['commit_batch'] - 1:
                data_ = self.data_dot_dict(commits, parents, devs, dev_com, branches,
                                           branches_com, files, com_files,
                                           methods, files_methods, com_methods)

                self.send_index_data(
                    {'project_conf': self.project, 'data': data_})

                commits, parents, devs, dev_com, branches, branches_com, files, com_files, methods, files_methods, com_methods = [
                ], [], [], [], [], [], [], [], [], [], []
                commit_index = 0
            else:
                commit_index += 1

        print('Driller finished in: \t', datetime.now() - start)



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" This module uses pydriller to search a repository
and indexes it in neo4j
"""
from datetime import datetime
from pydriller import RepositoryMining

import graphrepo.utils as utl
import graphrepo.drillers.batch_utils as b_utl
from graphrepo.drillers.drill_cache import DrillCache, DrillCacheSequential
from graphrepo.drillers.default import DefaultDriller
from graphrepo.logger import Logger

LG = Logger()


class CacheDriller(DefaultDriller):
    """CacheDriller class - parses a git repo and uses the models
    to index everything in Neo4j by storing all data on disk.
    """

    def drill_batch_cache_sequential(self, index=True):
        """Extracts all information from a git repository
        and it stores in in a disk cache
        :param index: optional, if True, the data is indexed in Neo4j
        :returns: cache with all data
        """
        start = datetime.now()
        print('Driller started at: \t', start)
        cache = DrillCacheSequential()
        for commit in \
            RepositoryMining(self.config.ct.repo,
                             since=self.config.ct.start_date,
                             to=self.config.ct.end_date).traverse_commits():
            timestamp = commit.author_date.timestamp()
            dev = utl.format_dev(commit, self.config.ct.index_developer_email)
            cache.append_cache('developers', dev)
            com = utl.format_commit(commit, self.config.ct.project_id)
            cache.append_cache('commits', com)
            cache.append_cache(
                'dev_commits',
                utl.format_author_commit(dev, com, timestamp))
            for parent in commit.parents:
                cache.append_cache('parents', utl.format_parent_commit(
                    com['hash'], parent, self.config.ct.project_id))
            for branch in commit.branches:
                br_ = utl.format_branch(branch, self.config.ct.project_id)
                cache.append_cache('branches', br_)
                cache.append_cache('branches_commits', utl.format_branch_commit(
                    br_['hash'], com['hash']))
            for file in commit.modifications:
                fl_ = utl.format_file(file, self.config.ct.project_id)
                cache.append_cache('files', fl_)
                cache.append_cache('commit_files', utl.format_commit_file(
                    com['hash'], fl_['hash'], file, timestamp))
                for method in file.changed_methods:
                    met = utl.format_method(
                        method, file, self.config.ct.project_id)
                    cache.append_cache('methods', met)
                    cache.append_cache(
                        'file_methods',
                        utl.format_file_method(fl_['hash'],
                                               met['hash']))
                    cache.append_cache('commit_methods',
                                       utl.format_commit_method(
                                           com['hash'],
                                           met['hash'],
                                           method,
                                           timestamp))
        print('Driller finished in: \t', datetime.now() - start)
        if index:
            self.index_batch(cache)
        return cache

    def index_batch(self, cache):
        """Indexes cached data to Neo4j
        :param cache: diskcache Cache or Index
        """
        try:
            self.config.check_config()
            self._check_connection()
            b_utl.index_cache(
                self.graph, cache, batch_size=self.config.ct.batch_size)
        except Exception as exc:
            LG.log_and_raise(exc)
        else:
            return

    def drill_batch_cache_all(self, index=True):
        """Extracts the information from a repository in memory
        and caches it after the extraction
        :param index: optional, if True, the data is indexed in Neo4j
        """
        data = self.drill_batch(index=False)
        cache = DrillCache(data)
        if index:
            self.index_batch(cache)
        return cache



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module mines developers and contains all related Neo4j queries"""

from graphrepo.miners.default import DefaultMiner
from graphrepo.miners.utils import format_commit_id_date as fcid


class DeveloperMiner(DefaultMiner):
    """This class holds queries for the Developer nodes"""

    def query(self, **kwargs):
        """Queries developers by any arguments given in kwargs
        For example kwargs can be {'hash': 'example-hash'} or
        {'email': 'example-email'}
        :param kwargs: any parameter and value, between hash, name or email
        :returns: list of nodes matched
        """
        return self.node_matcher.match("Developer", **kwargs)

    def get_commits(self, dev_hash, project_id=None,
                    start_date=None, end_date=None):
        """Returns all commits authored by a developer.
        Optionally, it also filters by project id
        :param dev_hash: developer unique identifier
        :param project_id: optional; if present the
          query returns the commits from a project
        :param start_date: optional timestamp; filter commits
          beginning with this date
        :param end_date: optional timestamp; filter commits
          untill this date
        :returns: list of commits
        """
        com_filter, where = fcid(project_id,
                                 start_date, end_date)
        cquery = """
        MATCH (d:Developer {{hash: "{0}"}})
              -[r:Author]->
              (c:Commit {1})
        {2}
        RETURN distinct c;
        """.format(dev_hash, com_filter, where)
        dt_ = self.graph.run(cquery)
        return [dict(x['c']) for x in dt_.data()]

    def get_files(self, dev_hash, project_id=None,
                  start_date=None, end_date=None):
        """Returns all files edited by a developer.
        Optionally it also filters by project_id
        :params dev_hash: developer unique identifier
        :params project_id: optional; if present the query
          returns the files from a specific project
        :param start_date: optional timestamp; filter files
          beginning with this date
        :param end_date: optional timestamp; filter files
          untill this date
        :returns: list of files
        """
        com_filter, where = fcid(project_id,
                                 start_date, end_date)
        fquery = """
        MATCH (d:Developer {{hash: "{0}"}})
              -[r:Author]->
              (c:Commit {1})
              -[UpdateFile]->
              (f: File)
        {2}
        RETURN collect(distinct f);
        """.format(dev_hash, com_filter, where)
        dt_ = self.graph.run(fquery)
        return [dict(x) for x in dt_.data()[0]['collect(distinct f)']]

    def get_files_updates(self, dev_hash, project_id=None,
                          start_date=None, end_date=None):
        """Returns all file update information (e.g. file complexity),
        for all files edited by a developer.
        Optionally it also filters by project_id
        :params dev_hash: developer unique identifier
        :params project_id: optional; if present the query
          returns the files from a specific project
        :param start_date: optional timestamp; filter files
          beginning with this date
        :param end_date: optional timestamp; filter files
          untill this date
        :returns: list of file updates
        """
        com_filter, where = fcid(project_id,
                                 start_date, end_date)
        fuquery = """
        MATCH (d:Developer {{hash: "{0}"}})
              -[r:Author]->
              (c:Commit {1})
              -[fu: UpdateFile]->
              (f: File)
        {2}
        RETURN distinct fu;
        """.format(dev_hash, com_filter, where)

        dt_ = self.graph.run(fuquery)
        return [dict(x['fu']) for x in dt_.data()]

    def get_methods(self, dev_hash, project_id=None,
                    start_date=None, end_date=None):
        """Returns all methods updated by a developer.
        Optionally it also filters by project_id
        :params dev_hash: developer unique identifier
        :params project_id: optional; if present the query
          returns the files from a specific project
        :param start_date: optional timestamp; filter files
          beginning with this date
        :param end_date: optional timestamp; filter files
          untill this date
        :returns: list of methods
        """
        com_filter, where = fcid(project_id,
                                 start_date, end_date)
        mquery = """
        MATCH (d:Developer {{hash: "{0}"}})
              -[r:Author]->
              (c:Commit {1})
              -[um: UpdateMethod]->
              (m: Method)
        {2}
        RETURN distinct m;
        """.format(dev_hash, com_filter, where)

        dt_ = self.graph.run(mquery)
        return [dict(x['m']) for x in dt_.data()]

    def get_method_updates(self, dev_hash, project_id=None,
                           start_date=None, end_date=None):
        """Returns all method update information, for all
        methods update by a developer.
        Optionally it also filters by project_id
        :params dev_hash: developer unique identifier
        :params project_id: optional; if present the query
          returns the files from a specific project
        :param start_date: optional timestamp; filter files
          beginning with this date
        :param end_date: optional timestamp; filter files
          untill this date
        :returns: list of method updates
        """
        com_filter, where = fcid(project_id,
                                 start_date, end_date)
        muquery = """
        MATCH (d:Developer {{hash: "{0}"}})
              -[r:Author]->
              (c:Commit {1})
              -[um: UpdateMethod]->
              ()
        {2}
        RETURN distinct um;
        """.format(dev_hash, com_filter, where)

        dt_ = self.graph.run(muquery)
        return [dict(x['um']) for x in dt_.data()]

    def get_all(self):
        return self.node_matcher.match("Developer")



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utils methods for GraphRepo"""
import json
import hashlib
from datetime import datetime
import yaml


class Dotdict(dict):
    """dot.notation access to dictionary attributes"""
    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def parse_config(path):
    with open(path, 'r') as ymlfile:
        conf = yaml.load(ymlfile, Loader=yaml.FullLoader)

    neo = conf['neo']
    project = conf['project']

    project['start_date'] = datetime.strptime(
        project['start_date'], '%d %B, %Y %H:%M') \
        if project['start_date'] else None
    project['end_date'] = datetime.strptime(
        project['end_date'], '%d %B, %Y %H:%M') \
        if project['end_date'] else None

    return neo, project


def save_json(path, data):
    with open(path, 'w') as outfile:
        json.dump(data, outfile)


def load_json(path):
    with open(path) as json_file:
        return json.load(json_file)


def get_file_hash(file, project_id=None, use_new_path=False):
    name = ''
    if not file.old_path and file.new_path:
        name = name+file.new_path
    elif file.old_path and not file.new_path:
        name = name+file.old_path
    elif file.old_path and file.new_path:
        if use_new_path:
            name = name+file.new_path
        else:
            name = name + file.old_path

    name = name+file.filename
    name = project_id + name if project_id else name
    return hashlib.sha224(str(name).encode('utf-8')).hexdigest()


def get_method_type(method, m_before, m_current):
    if method.name in m_before and method.name not in m_current:
        return "DELETE"
    elif method.name in m_before and method.name in m_current:
        return "MODIFY"
    else:
        return "ADD"


def get_method_hash(method, file, project_id=None):
    fhash = get_file_hash(file, project_id)
    _fmname = fhash + "_" + method.name
    _fmname = project_id + _fmname if project_id else _fmname
    return hashlib.sha224(_fmname.encode('utf-8')).hexdigest()


def get_author_hash(email):
    return hashlib.sha224(email.encode('utf-8')).hexdigest()


def format_dev(dev, index_email=True):
    return {
        'name': dev.author.name,
        'email': dev.author.email if index_email else '',
        'hash': get_author_hash(dev.author.email)
    }


def get_commit_hash(chash, project_id):
    return hashlib.sha224(str(project_id + chash).encode('utf-8')).hexdigest()


def format_commit(com, project_id):
    return {
        'hash': get_commit_hash(com.hash, project_id),
        'commit_hash': com.hash,
        'message': com.msg,
        'is_merge': 1 if com.merge else 0,
        'timestamp': com.author_date.timestamp(),
        'project_id': project_id,
        'dmm_unit_complexity': com.dmm_unit_complexity if com.dmm_unit_complexity else -1,
        'dmm_unit_interfacing': com.dmm_unit_interfacing if com.dmm_unit_interfacing else -1,
        'dmm_unit_size': com.dmm_unit_size if com.dmm_unit_size else -1,
    }


def format_parent_commit(c_hash, parent_hash, project_id=None):
    return {
        'child_hash': c_hash,
        'parent_hash': get_commit_hash(parent_hash, project_id)
    }


def format_branch(name, project_id):
    return {
        'hash':  hashlib.sha224(str(project_id+name).encode('utf-8')).hexdigest(),
        'project_id': project_id,
        'name': name
    }


def format_author_commit(dev, com, timestamp):
    return {'commit_hash': com['hash'],
            'author_hash': dev['hash'],
            'timestamp': timestamp,
            }


def format_branch_commit(bhash, chash):
    return {'branch_hash': bhash,
            'commit_hash': chash
            }


def format_file(file, project_id):
    return {
        'hash': get_file_hash(file, project_id),
        'merge_hash': get_file_hash(file, project_id, use_new_path=True),
        'name': file.filename,
        'project_id': project_id,
        'type': '.' + file.filename.split('.')[-1:][0]
    }


def format_commit_file(c_hash, f_hash, file, timestamp, index_code=True):
    dt_ = {'commit_hash': c_hash, 'file_hash': f_hash,
           'attributes': {
               'timestamp': timestamp,
               'old_path': file.old_path if file.old_path else '',
               'path': file.new_path if file.new_path else '',
               'source_code': '',
               'source_code_before': '',
               'diff': file.diff,
               'nloc': file.nloc if file.nloc else -1,
               'complexity': file.complexity if file.complexity else -1,
               'token_count': file.token_count if file.token_count else -1,
               'added': file.added,
               'removed': file.removed,
               'type': file.change_type.name}}

    if index_code:
        dt_['attributes']['source_code'] = str(
            file.source_code) if file.source_code else '',
        dt_['attributes']['source_code_before'] = str(
            file.source_code_before) if file.source_code_before else '',

    return dt_


def format_commit_method(c_hash, m_hash, met, timestamp):
    return {
        'commit_hash': c_hash,
        'method_hash': m_hash,
        'attributes': {
            'timestamp': timestamp,
            'long_name': met.long_name,
            'parameters': met.parameters,
            'complexity': met.complexity,
            'nloc': met.nloc,
            'fan_in': met.fan_in,
            'fan_out': met.fan_out,
            'general_fan_out': met.general_fan_out,
            'length': met.length,
            'token_count': met.token_count,
            'start_line': met.start_line,
            'end_line': met.end_line}}


def format_method(met, fille, project_id):
    return {
        'hash': get_method_hash(met, fille, project_id),
        'name': met.name,
        'file_name': met.filename,
        'project_id': project_id}


def format_file_method(f_hash, m_hash):
    return {'file_hash': f_hash, 'method_hash': m_hash}



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module stores all config constants. It is a singleton
because it is used across several modules inside the app"""

from graphrepo.singleton import Singleton
from graphrepo.utils import Dotdict


class Config(metaclass=Singleton):
    """This class contains all config flags"""
    ct = {}

    def configure(self, **kwargs):
        """Stores configuration contants, parsed
        from yaml config file
        :param kwargs: keys and values from config
        """
        self.ct = Dotdict(kwargs)

    def check_config(self):
        """Checks if the config properties are set and
        raises ValueError if any value misses"""

        if not self.ct.db_url or not self.ct.port \
                or not self.ct.db_user or not self.ct.db_pwd:
            raise ValueError("Neo4j configuartion is invalid.")




# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Default Parent class for drillers
"""
from abc import abstractmethod
from datetime import datetime
from py2neo import Graph
from pydriller import RepositoryMining

import graphrepo.utils as utl
import graphrepo.drillers.db_init as db_init
from graphrepo.config import Config
from graphrepo.logger import Logger
LG = Logger()


class DefaultDriller():
    """DefaultDriller class - parses a git repo and uses the models
    to index everything in Neo4j.
    """

    def __init__(self, config_path):
        """Initializes the properties of this class
        :param config_path: path to yml config file
        """
        try:
            if not config_path:
                raise FileNotFoundError
            neo, project = utl.parse_config(config_path)
            self.config = Config()
            self.graph = None
            self.config.configure(**neo, **project)
            self._connect()
        except Exception as exc:
            LG.log_and_raise(exc)

    def _connect(self):
        """Instantiates the connection to Neo4j and stores
        the graph internally.
        Throws exception if the connection can not pe realized
        """
        try:
            self.graph = Graph(host=self.config.ct.db_url,
                               user=self.config.ct.db_user,
                               password=self.config.ct.db_pwd,
                               port=self.config.ct.port)
        except Exception as exc:
            LG.log_and_raise(exc)

    def _check_connection(self):
        """Checks if there is a db connection and raises
        ReferenceError if not.
        """
        try:
            self._connect()
        except:
            raise ReferenceError("There is no valid "
                                 "database connection. Please "
                                 "configure and connect first.")

    def init_db(self):
        """Runs initialization of a database; creates
        constraints and indexes"""
        try:
            self._check_connection()
            db_init.create_hash_constraints(self.graph)
            db_init.create_indices(self.graph, hash_index=False)
        except Exception as e:
            raise e

    def clean(self):
        """Removes all data in a graph
        """
        try:
            self.config.check_config()
            self._check_connection()

            self.graph.run("MATCH (n) OPTIONAL MATCH (n)-[r]-() DELETE n,r")
        except Exception as exc:
            LG.log_and_raise(exc)

    def drill_batch(self, index=True, save_path=None):
        """Extracts data from a software repository, with the option
        of saving it on diks and indexing it in Neo4j
        :param index: optional, if True, the data is indexed in Neo4j
        :param save_path: optional, if given, the data is stored on dik
        :returns: dictionary with all data
        """
        start = datetime.now()
        print('Driller started at: \t', start)
        commits, parents, devs, dev_com, branches,\
            branches_com, files, com_files, \
            methods, files_methods, com_methods = \
            [], [], [], [], [], [], [], [], [], [], []
        for commit in \
            RepositoryMining(self.config.ct.repo,
                             since=self.config.ct.start_date,
                             to=self.config.ct.end_date).traverse_commits():
            self.drill_commit(commit, commits, parents, devs, dev_com, branches,
                              branches_com, files, com_files,
                              methods, files_methods, com_methods)

        data_ = self.data_dot_dict(commits, parents, devs, dev_com, branches,
                                   branches_com, files, com_files,
                                   methods, files_methods, com_methods)

        print('Driller finished in: \t', datetime.now() - start)

        if save_path:
            utl.save_json(save_path, data_)
        if index:
            self.index_batch(**data_)
        return data_

    def drill_commit(self, commit, commits, parents, devs, dev_com, branches,
                     branches_com, files, com_files,
                     methods, files_methods, com_methods):
        """Helper method - works with pass by reference"""
        timestamp = commit.author_date.timestamp()
        dev = utl.format_dev(commit, self.config.ct.index_developer_email)
        devs.append(dev)
        com = utl.format_commit(commit, self.config.ct.project_id)
        commits.append(com)
        dev_com.append(utl.format_author_commit(dev, com, timestamp))
        for parent in commit.parents:
            parents.append(utl.format_parent_commit(
                com['hash'], parent, self.config.ct.project_id))
        for branch in commit.branches:
            br_ = utl.format_branch(branch, self.config.ct.project_id)
            branches.append(br_)
            branches_com.append(
                utl.format_branch_commit(br_['hash'], com['hash']))
        for file in commit.modifications:
            fl_ = utl.format_file(file, self.config.ct.project_id)
            files.append(fl_)
            com_files.append(utl.format_commit_file(
                com['hash'], fl_['hash'], file,
                timestamp, self.config.ct.index_code))
            for method in file.changed_methods:
                met = utl.format_method(
                    method, file, self.config.ct.project_id)
                methods.append(met)
                files_methods.append(
                    utl.format_file_method(fl_['hash'], met['hash'])
                )
                com_methods.append(
                    utl.format_commit_method(com['hash'], met['hash'],
                                             method, timestamp))

    def data_dot_dict(self, commits, parents, devs, dev_com, branches,
                      branches_com, files, com_files,
                      methods, files_methods, com_methods):
        """Helper method"""
        return utl.Dotdict({'commits': commits,
                            'parents': parents,
                            'developers': devs,
                            'dev_commits': dev_com,
                            'branches': branches,
                            'branches_commits': branches_com,
                            'files': files,
                            'commit_files': com_files,
                            'methods': methods,
                            'file_methods': files_methods,
                            'commit_methods': com_methods})

    @abstractmethod
    def index_batch(self):
        """Abstract index batch driller method
        """
        raise NotImplementedError



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from graphrepo.drillers.cache_driller import CacheDriller


class TestCacheDriller:
    def test_indexing(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = CacheDriller(os.path.join(folder, 'cnfg_init.yml'))
        test_driller.drill_batch_cache_sequential()
        records = [r for r in test_driller.graph.run(
            "MATCH(n) RETURN n")]
        assert len(records) == 22

        test_driller.clean()

    def test_drill_batch_cache(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = CacheDriller(os.path.join(folder, 'cnfg_init.yml'))
        test_driller.drill_batch_cache_all()
        records = [r for r in test_driller.graph.run(
            "MATCH(n) RETURN n")]
        assert len(records) == 22

        test_driller.clean()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module mines files and contains all related Neo4j queries"""

from graphrepo.miners.default import DefaultMiner


class FileMiner(DefaultMiner):
    """This clas holds queries for the File nodes"""

    def query(self, **kwargs):
        """Searches for a file using the arguments in kwargs.
        If no kwargs are given it returns the first file found
        """
        return self.node_matcher.match("File", **kwargs).first()

    def get_all(self):
        """Returns all node of type File
        :return: list of files
        """
        return self.node_matcher.match("File")

    def get_change_history(self, file_hash):
        """Returns all updated relationships
        :param file_hash: a string, unique identifier for file
        :param dic: optional; boolean for converting data to dictionary
          or returning it as py2neo records - the py2neo raw
          records can be used in mappers
        :return: list of update file relationships
        """
        query = """MATCH ()-[r:UpdateFile]->(f:File {{hash: "{0}"}})
        return distinct r
        """.format(file_hash)
        dt_ = self.graph.run(query)
        return [dict(x['r']) for x in dt_.data()]

    def get_current_methods(self, file_hash):
        """Returns all current methods
        :param file_hash: a string, unique identifier for file
        :param dic: optional; boolean for converting data to dictionary
          or returning it as py2neo records - the py2neo raw
          records can be used in mappers
        :return: list of methods
        """
        query = """MATCH (f:File {{hash: "{0}"}})-[r:Method]->(m:Method)
        return distinct m
        """.format(file_hash)
        dt_ = self.graph.run(query)
        return [dict(x['m']) for x in dt_.data()]

    def get_past_methods(self, file):
        """Returns methods that were removed from the file
          :param file: Py2Neo File object
          :returrn: list of Method objects
          """
        # return [rel.end_node
        #         for rel in self.graph.match([file, None], "HadMethod")]



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import os

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.drillers.driller import Driller
from graphrepo.miners.developer import DeveloperMiner


class TestDevMiner:
    def test_gets(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        st_date = datetime.strptime(
            "14 May, 2020 00:00", '%d %B, %Y %H:%M').timestamp()
        end_date = datetime.strptime(
            "15 May, 2020 02:00", '%d %B, %Y %H:%M').timestamp()

        n_matcher = NodeMatcher(test_driller.graph)
        r_matcher = RelationshipMatcher(test_driller.graph)

        dev_miner = DeveloperMiner(test_driller.graph, n_matcher, r_matcher)

        all_devs = dev_miner.get_all()
        assert len(all_devs) == 2

        all_commits = dev_miner.get_commits(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb")
        assert len(all_commits) == 7

        all_com_id = dev_miner.get_commits(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb",
            project_id=test_driller.config.ct.project_id
        )
        assert len(all_com_id) == 7

        all_com_id_dates = dev_miner.get_commits(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb",
            project_id=test_driller.config.ct.project_id,
            start_date=st_date,
            end_date=end_date
        )
        assert len(all_com_id_dates) == 7

        all_files = dev_miner.get_files(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb"
        )
        assert len(all_files) == 6

        all_files_id_dates = dev_miner.get_files(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb",
            project_id=test_driller.config.ct.project_id,
            start_date=st_date,
            end_date=end_date
        )
        assert len(all_files_id_dates) == 6

        files_updates = dev_miner.get_files_updates(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb"
        )
        assert len(files_updates) == 9

        files_updates_id_dates = dev_miner.get_files_updates(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb",
            project_id=test_driller.config.ct.project_id,
            start_date=st_date,
            end_date=end_date
        )
        assert len(files_updates_id_dates) == 9

        all_methods = dev_miner.get_methods(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb"
        )
        assert len(all_methods) == 5

        all_methods_id_dates = dev_miner.get_methods(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb",
            project_id=test_driller.config.ct.project_id,
            start_date=st_date,
            end_date=end_date
        )
        assert len(all_methods_id_dates) == 5

        method_updates = dev_miner.get_method_updates(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb"
        )
        assert len(method_updates) == 9

        method_updates_id_dates = dev_miner.get_method_updates(
            dev_hash="bb1a1830d2f4f4d13151827aa1072ed43bd8738a139da332e1ee3ddb",
            project_id=test_driller.config.ct.project_id,
            start_date=st_date,
            end_date=end_date
        )
        assert len(method_updates_id_dates) == 9

        test_driller.clean()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


###
# This file assumes the project from the config file was already indexed
###
import argparse
import os
import pandas as pd
import plotly.express as px

from datetime import datetime
from graphrepo.miners import MineManager
from graphrepo.utils import parse_config


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config', default='examples/configs/pydriller.yml', type=str)
    return parser.parse_args()


def main():
    args = parse_args()
    mine_manager = MineManager(config_path=args.config)
    files = mine_manager.dev_miner.get_files(
        "6cf1f138e29c1bf82810ad0b73012302e0d20c2f76a24e3b225017b0",
        mine_manager.config.ct.project_id
    )
    print(len(files), ' files')

    file_updates = mine_manager.dev_miner.get_files_updates(
        "6cf1f138e29c1bf82810ad0b73012302e0d20c2f76a24e3b225017b0",
        mine_manager.config.ct.project_id
    )
    print(len(file_updates), ' file updates')

    methods = mine_manager.dev_miner.get_methods(
        "6cf1f138e29c1bf82810ad0b73012302e0d20c2f76a24e3b225017b0",
        mine_manager.config.ct.project_id
    )
    print(len(methods), ' methods')

    method_updates = mine_manager.dev_miner.get_method_updates(
        "6cf1f138e29c1bf82810ad0b73012302e0d20c2f76a24e3b225017b0",
        mine_manager.config.ct.project_id
    )
    print(len(method_updates), ' method updates')


if __name__ == '__main__':
    main()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module initializes and configures all miners"""
from py2neo import Graph, NodeMatcher, RelationshipMatcher
import graphrepo.utils as utl
from graphrepo.config import Config
from graphrepo.logger import Logger
from graphrepo.singleton import Singleton
from graphrepo import miners


LG = Logger()


class MineManager(metaclass=Singleton):
    """MineManageer class - This class manages custom
    miners. At the moment we instantiate all miners,
    but other managers which handle different 'teams of miners'
    can be created.
    """

    def __init__(self, config_path):
        """Initializes the properties of this class"""
        self.commit_miner, self.dev_miner, \
            self.file_miner, self.method_miner = None, None, None, None
        try:
            if not config_path:
                raise FileNotFoundError
            neo, project = utl.parse_config(config_path)
            self.config = Config()
            self.config.configure(**neo, **project)
            self.graph = None
            self.node_matcher = None
            self.rel_matcher = None
            self.connect()
        except Exception as exc:
            LG.log_and_raise(exc)

    def connect(self):
        """Instantiates the connection to Neo4j and stores
        the graph internally.
        Throws exception if the connection can not pe realized
        """
        try:
            self.graph = Graph(host=self.config.ct.db_url,
                               user=self.config.ct.db_user,
                               password=self.config.ct.db_pwd,
                               http_port=self.config.ct.port)
            self.node_matcher = NodeMatcher(self.graph)
            self.rel_matcher = RelationshipMatcher(self.graph)
            self.init_miners()
        except Exception as exc:
            LG.log_and_raise(exc)

    def check_connection(self):
        """Checks if there is a db connection and raises
        ReferenceError if not.
        """
        try:
            self.connect()
        except:
            raise ReferenceError("There is no valid "
                                 "database connection. Please "
                                 "configure and connect first.")

    def init_miners(self):
        """Initializes all miners"""
        try:
            # TODO: Parse this automatically?
            self.commit_miner = miners.CommitMiner(
                graph=self.graph,
                node_matcher=self.node_matcher,
                rel_matcher=self.rel_matcher)
            self.dev_miner = \
                miners.DeveloperMiner(graph=self.graph,
                                      node_matcher=self.node_matcher,
                                      rel_matcher=self.rel_matcher)
            self.file_miner = \
                miners.FileMiner(graph=self.graph,
                                 node_matcher=self.node_matcher,
                                 rel_matcher=self.rel_matcher)
            self.method_miner = \
                miners.MethodMiner(graph=self.graph,
                                   node_matcher=self.node_matcher,
                                   rel_matcher=self.rel_matcher)

        except Exception as exc:
            LG.log_and_raise(exc)
        else:
            return

    def get_all_data(self):
        """Returns all nodes and relationships from Neo4j
        :returns:  a tuple with two arrays: the first with nodes,
            the second with relationships
        """
        nodes = self.node_matcher.match()
        rels = self.rel_matcher.match()

        return list(nodes), list(rels)



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Singleton metaclass"""


class Singleton(type):
    """
    Define an Instance operation that lets clients access its unique
    instance.
    """

    def __init__(cls, name, bases, attrs, *args, **kwargs):
        super().__init__(name, bases, attrs)
        cls._instance = None

    def __call__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__call__(*args, **kwargs)
        return cls._instance



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module mines commits and contains all related Neo4j queries"""

from graphrepo.miners.default import DefaultMiner
from graphrepo.miners.utils import format_commit_id_date


class CommitMiner(DefaultMiner):
    """This class holds queries for commits"""

    def query(self, **kwargs):
        """Queries commits by any arguments given in kwargs
        For example kwargs can be {'hash': 'example-hash'}
        :param kwargs: any parameter and value, between hash, name or email
        :returns: list of commit nodes matched
        """
        com_ = self.node_matcher.match("Commit", **kwargs)
        return [dict(x) for x in com_]

    def get_between_dates(self, start_date, end_date,
                          project_id=None):
        """Returns all commits between start and end date
        :param start_date: timestamp, start date
        :param end_date: timestamp, end date
        :param project_id: optional; if given only the commits from a project
          are returned
        :returns: list of commitss
        """
        com_filter, where = format_commit_id_date(
            project_id, start_date, end_date)
        query = """
        MATCH (c: Commit {0})
        {1}
        RETURN distinct c
        """.format(com_filter, where)
        dt_ = self.graph.run(query)
        return [dict(x['c']) for x in dt_.data()]

    def get_all(self,):
        """Returns all commits
        :returns: list of commit nodes
        """
        com_ = self.node_matcher.match("Commit")
        return [dict(x) for x in com_]

    def get_commit_files(self, commit_hash):
        """Returns the files updated in a commit
        :param commit_hash: optional; if given, it will
          return the data only for one commit
        :returns: list of commit files
        """
        query = """
          MATCH (c:Commit {{hash: "{0}"}})
          -[UpdateFile]->(f:File)
          return distinct f
          """.format(commit_hash)
        files_ = self.graph.run(query)
        return [x['f'] for x in files_.data()]

    def get_commit_file_updates(self, commit_hash):
        """Returns the updates a commit made to files (UpdateFile rel)
        :param commit_hash: optional; if given, it will
          return the data only for one commit
        :returns: list of
        """
        query = """
          MATCH (c:Commit {{hash: "{0}"}})
          -[f: UpdateFile]->(fu:File)
          return distinct f
          """.format(commit_hash)
        files_ = self.graph.run(query)
        return [x['f'] for x in files_.data()]

    def get_commit_methods(self, commit_hash=None):
        """Returns the methods updated in a commit
        :param commit_hash: optional; if given, it will
          return the data only for one commit
        """
        query = """
          MATCH (c:Commit {{hash: "{0}"}})
          -[UpdateMethod]->(m:Method)
          return distinct m
          """.format(commit_hash)
        files_ = self.graph.run(query)
        return [x['m'] for x in files_.data()]

    def get_commit_method_updates(self, commit_hash=None):
        """Returns the updatemethod relationships from a commit
        :param commit_hash: optional; if given,
          it will return the data only for one commit
        :param dic: optional, boolean for ocnverting the data to dictionaries
        """
        query = """
          MATCH (c:Commit {{hash: "{0}"}})
          -[m:UpdateMethod]->(mu:Method)
          return distinct m
          """.format(commit_hash)
        files_ = self.graph.run(query)
        return [x['m'] for x in files_.data()]



"""This module is the wild wild west of batch indexing :-)
In contains all Neo4j queries for indexing the data in batches.
More documentation will follow soon.
"""
from datetime import datetime


def batch(iterable, n=1):
    l = len(iterable)
    for ndx in range(0, l, n):
        yield iterable[ndx:min(ndx + n, l)]


def index_commits(graph, commits, batch_size=100):
    query = """
    UNWIND {commits} AS c
    MERGE (nc :Commit { hash: c.hash})
      ON CREATE SET
                nc = c
      ON MATCH SET
                nc = c
    """
    for b in batch(commits, batch_size):
        graph.run(query, commits=b)


def index_parent_commits(graph, parents, batch_size=100):
    query = """
    UNWIND {ac} AS a
    MATCH (x:Commit),(y:Commit)
    WHERE x.hash = a.parent_hash AND y.hash = a.child_hash
    MERGE (x)-[r:Parent{}]->(y)
    """
    for b in batch(parents, batch_size):
        graph.run(query, ac=b)


def index_authors(graph, authors, batch_size=100):
    query = """
    UNWIND {authors} AS a
    MERGE (nd:Developer { hash: a.hash})
      ON CREATE SET nd = a
      ON MATCH SET nd = a
    """
    for b in batch(authors, batch_size):
        graph.run(query, authors=b)


def index_branches(graph, branches, batch_size=100):
    query = """
    UNWIND {branches} AS a
    MERGE (nb:Branch { hash: a.hash})
      ON CREATE SET nb = a
      ON MATCH SET nb = a
    """
    for b in batch(branches, batch_size):
        graph.run(query, branches=b)


def index_branch_commits(graph, bc, batch_size=100):
    query = """
    UNWIND {ac} AS a
    MATCH (x:Branch),(y:Commit)
    WHERE x.hash = a.branch_hash AND y.hash = a.commit_hash
    MERGE (x)-[r:BranchCommit{}]->(y)
    """
    for b in batch(bc, batch_size):
        graph.run(query, ac=b)


def index_files(graph, files, batch_size=100):
    query = """
    UNWIND {files} AS f
    MERGE (nf:File { hash: f.hash})
      ON CREATE SET nf = f
      ON MATCH SET nf = f
    """
    for b in batch(files, batch_size):
        graph.run(query, files=b)


def index_methods(graph, methods, batch_size=100):
    query = """
    UNWIND {methods} AS f
    MERGE (nm:Method { hash: f.hash})
      ON CREATE SET nm = f
      ON MATCH SET nm = f
    """

    for b in batch(methods, batch_size):
        graph.run(query, methods=b)


def index_author_commits(graph, ac, batch_size=100):
    query = """
    UNWIND {ac} AS a
    MATCH (x:Developer),(y:Commit)
    WHERE x.hash = a.author_hash AND y.hash = a.commit_hash
    MERGE (x)-[r:Author{timestamp: a.timestamp}]->(y)
    """
    for b in batch(ac, batch_size):
        graph.run(query, ac=b)


def index_commit_files(graph, cf, batch_size=100):
    query = """
    UNWIND {cf} AS a
    MATCH (x:Commit),(y:File)
    WHERE x.hash = a.commit_hash AND y.hash = a.file_hash
    MERGE (x)-[r:UpdateFile{}]->(y)
    ON CREATE SET r=a['attributes']
    """
    for i, b in enumerate(batch(cf, batch_size)):
        graph.run(query, cf=b)


def index_file_methods(graph, cf, batch_size=100):
    query = """
    UNWIND {cf} AS a
    MATCH (x:File),(y:Method)
    WHERE x.hash = a.file_hash AND y.hash = a.method_hash
    MERGE (x)-[r:Method{}]->(y)
    """
    for b in batch(cf, batch_size):
        graph.run(query, cf=b)


def index_commit_method(graph, cm, batch_size=100):
    query = """
    UNWIND {cf} AS a
    MATCH (x:Commit),(y:Method)
    WHERE x.hash = a.commit_hash AND y.hash = a.method_hash
    MERGE (x)-[r:UpdateMethod]->(y)
    ON CREATE SET r=a['attributes']
    """
    for i, b in enumerate(batch(cm, batch_size)):
        graph.run(query, cf=b)


def create_index_authors(graph):
    query = """
    CREATE INDEX ON :Developer(hash)
    """
    graph.run(query)


def create_index_commits(graph, hash=True):
    if hash:
        hash_q = """
        CREATE INDEX ON :Commit(hash)
        """
        graph.run(hash_q)

    pid_q = """
    CREATE INDEX ON :Commit(project_id)
    """

    graph.run(pid_q)


def create_index_branches(graph, hash=True):
    if hash:
        hash_q = """
      CREATE INDEX ON :Branch(hash)
      """
        graph.run(hash_q)

    pid_q = """
    CREATE INDEX ON :Branch(project_id)
    """
    graph.run(pid_q)


def create_index_files(graph, hash=True):
    if hash:
        hash_q = """
        CREATE INDEX ON :File(hash)
        """
        graph.run(hash_q)

    pid_q = """
    CREATE INDEX ON :File(project_id)
    """
    graph.run(pid_q)


def create_index_methods(graph, hash=True):
    if hash:
        hash_q = """
        CREATE INDEX ON :Method(hash)
        """
        graph.run(hash_q)

    pid_q = """
    CREATE INDEX ON :Method(project_id)
    """
    graph.run(pid_q)


def merge_files(graph):
  query = """
    MATCH (n1:File),(n2:File)
    WHERE n1.merge_hash = n2.merge_hash  and id(n1) < id(n2)
    WITH [n1,n2] as ns
    order by id(ns[1])
    CALL apoc.refactor.mergeNodes(ns, {properties: 'overwrite', mergeRels:true}) YIELD node
    MATCH (f:File {hash: node.hash}) -[]->(mf:Method) WITH DISTINCT f, mf
    with collect({hash: mf.hash, new_hash: f.hash}) as allRows
    unwind allRows as row
    match (mu: Method {hash: row.hash})
    SET mu.merge_hash = row.new_hash"""
  graph.run(query)

def merge_methods(graph):
  query = """
  MATCH (n1:Method),(n2:Method)
  WHERE n1.file_name = n2.file_name and n1.name = n2.name and n1.project_id = n2.project_id and n1.merge_hash = n2.merge_hash and id(n1) < id(n2)
  WITH [n1,n2] as ns
  order by id(ns[1])
  CALL apoc.refactor.mergeNodes(ns, {properties: 'overwrite', mergeRels:true}) YIELD node
  return node
  """
  graph.run(query)

def index_all(graph, developers, commits, parents, dev_commits, branches,
              branches_commits, files, commit_files, methods, file_methods,
              commit_methods, batch_size=100):

    total = datetime.now()

    developers = list({v['hash']: v for v in developers}.values())
    print('Indexing ', len(developers), ' authors')
    start = datetime.now()
    index_authors(graph, developers, batch_size)
    print('Indexed authors in: \t', datetime.now()-start)

    print('Indexing ', len(commits), ' commits')
    start = datetime.now()
    index_commits(graph, commits, batch_size)
    print('Indexed commits in: \t', datetime.now()-start)

    branches = list({v['hash']: v for v in branches}.values())
    branches_commits = list({str(i): i for i in branches_commits}.values())
    print('Indexing ', len(branches), ' branches')
    start = datetime.now()
    index_branches(graph, branches, batch_size)
    index_branch_commits(graph, branches_commits, batch_size)
    print('Indexed branches in: \t', datetime.now()-start)

    files = list({v['hash']: v for v in files}.values())
    print('Indexing ', len(files), ' files')
    start = datetime.now()
    index_files(graph, files, batch_size)
    print('Indexed files in: \t', datetime.now()-start)

    methods = list({v['hash']: v for v in methods}.values())
    print('Indexing ', len(methods), ' methods')
    start = datetime.now()
    index_methods(graph, methods, batch_size)
    print('Indexed methods in: \t', datetime.now()-start)

    parents = list({str(i): i for i in parents}.values())
    print('Indexing ', len(parents), ' parent commits')
    start = datetime.now()
    index_parent_commits(graph, parents, batch_size)
    print('Indexed commits in: \t', datetime.now()-start)

    print('Indexing ', len(dev_commits), ' author_commits')
    start = datetime.now()
    index_author_commits(graph, dev_commits, batch_size)
    print('Indexed author_commits in: \t', datetime.now()-start)

    file_methods = list({str(i): i for i in file_methods}.values())
    print('Indexings ', len(file_methods), ' file_methods')
    start = datetime.now()
    index_file_methods(graph, file_methods, batch_size)
    print('Indexed file_methods in: \t', datetime.now()-start)

    print('Indexing ', len(commit_methods), ' commit_methods')
    start = datetime.now()
    index_commit_method(graph, commit_methods, batch_size)
    print('Indexed commit_methods in: \t', datetime.now()-start)

    print('Indexing ', len(commit_files), ' commit_files')
    start = datetime.now()
    index_commit_files(graph, commit_files, batch_size)
    print('Indexed commit_files in: \t', datetime.now()-start)

    print('Merging moved files and methods')
    start = datetime.now()
    merge_files(graph)
    merge_methods(graph)
    print('Merged files and methods in \t', datetime.now()-start)

    print('Indexing took: \t', datetime.now()-total)


def index_cache(graph, cache, batch_size=100):
    total = datetime.now()
    index_authors(graph, list(
        {v['hash']: v for v in cache.data['developers']}.values()), batch_size)
    index_commits(graph, cache.data['commits'], batch_size)
    index_branches(graph, list(
        {v['hash']: v for v in cache.data['branches']}.values()), batch_size)
    index_branch_commits(graph,  list(
        {str(i): i for i in cache.data['branches_commits']}.values()), batch_size)
    index_files(graph, list(
        {v['hash']: v for v in cache.data['files']}.values()), batch_size)
    index_methods(graph, list(
        {v['hash']: v for v in cache.data['methods']}.values()), batch_size)
    index_parent_commits(graph, list(
        {str(i): i for i in cache.data['parents']}.values()), batch_size)
    index_author_commits(graph, cache.data['dev_commits'], batch_size)
    index_file_methods(graph, list(
        {str(i): i for i in cache.data['file_methods']}.values()), batch_size)
    index_commit_method(graph, cache.data['commit_methods'], batch_size)
    index_commit_files(graph, cache.data['commit_files'], batch_size)
    merge_files(graph)
    merge_methods(graph)
    print('Indexing took: \t', datetime.now()-total)



# def delete_all():
#     # get total #of nodes
#     res = session.run("MATCH(n) RETURN COUNT(*) AS n")
#     total_nodes = 0
#     for item in res:
#         total_nodes = item["n"]
#     print("\n Existing nodes in db:", total_nodes)

#     # get total #of relationships
#     res1 = session.run("MATCH (n)-[r]->() RETURN COUNT(r) as r")
#     total_rels = 0
#     for item in res1:
#         total_rels = item["r"]
#     print("\n Existing relationships in db:", total_rels)

#     # delete all nodes in batches (for faster deletion)
#     while total_nodes > 0:
#         res = session.run(
#             "MATCH(n) WITH n LIMIT 10000 DETACH DELETE n RETURN COUNT(n) AS count")
#         count = 0
#         for item in res:
#             count = item["count"]  # updates deleeted node count here
#         total_nodes = total_nodes-count
#     print("\n #of nodes in db after deletion completed = ", total_nodes)


# start = time.time()
# delete_all()
# print("\n Pre cleanup time (sec): ", time.time()-start)

# for prot in fileList:
#     print("\n\n", prot)
#     if os.path.exists(prot+"_AllCCs_maxDist11.csv"):
#         print("\n Already Processed.")
#         continue
#     start = time.time()
#     delete_all()
#     pre_time = time.time()-start
#     print("\n Pre cleanup time (sec): ", pre_time)

#     # Database preparation
#     session.run("CREATE INDEX ON :MyNode(Name)")

#     # 1. Create graph
#     start = time.time()
#     session.run("USING PERIODIC COMMIT "
#                 "LOAD CSV FROM 'file:///'+{prot}+'_conflict_resolved.txt' AS line "
#                 "MERGE (n:MyNode {Name:line[0]}) "
#                 "MERGE (m:MyNode {Name:line[1]}) "
#                 "MERGE (n) -[:TO {dist:line[2]}] -> (m) ", prot=prot)

#     end = time.time()
#     step1_time = end - start
#     print("\n Step 1 time (in sec) = ", end-start)

#     # 2 find CCs
#     start = time.time()
#     result = session.run("CALL algo.unionFind.stream('MyNode', 'TO', {graph:'huge'}) "
#                          "YIELD nodeId,setId "
#                          "MATCH (n) "
#                          "WHERE id(n)=nodeId "
#                          "WITH setId,collect(nodeId) as nodes, collect(n.Name) as labels,count(*) as size_of_component "
#                          "ORDER BY size_of_component DESC "
#                          "RETURN setId as componentId,size_of_component,labels as connectedTSRkeys ")
#     end = time.time()
#     step2_time = end - start
#     print("\n Step 2 time (in sec) = ", end-start)
#    # 3. save result
#     start = time.time()
#     # newline='' <- to avoid blank line between two rows
#     with open(prot+"_AllCCs_maxDist11.csv", "w") as csvfile:
#         writer = csv.writer(csvfile, delimiter=',')
#         writer.writerow(
#             ['componentId', 'size_of_component', 'connectedTSRkeys'])
#         for record in result:
#             record = str(record)[:-1].replace(", ",
#                                               ",").replace("'", "").split()
#             print("\n", record[1], record[2], record[3])
#             writer.writerow([record[1].split("=")[1], record[2].split("=")[
#                             1], record[3].split("=")[1]])
#     end = time.time()
#     step3_time = end - start
#     print("\n Step 3 time (in sec) = ", end-start)

#     # 4. delete graph
#     start = time.time()
#     delete_all()
#     end = time.time()
#     post_time = end - start
#     print("\n Post cleanup time (in sec) = ", end-start)

#     print("\n Total time = ", pre_time+step1_time +
#           step2_time+step3_time+post_time)

# driver.close()



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module plots the method complexity evolution over time, for a file"""

###
# This file assumes the project from the config file was already indexed
###
import argparse
import pandas as pd
import plotly.express as px

from graphrepo.miners import MineManager

from datetime import datetime


def parse_args():
    """Parse args"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config', default='examples/configs/pydriller.yml', type=str)
    parser.add_argument('--plot', default=False, type=bool)
    return parser.parse_args()


def main():
    """Main"""
    args = parse_args()

    file_query = {
        'hash': 'e2eb7bf414cebe68f46fa88e4abe9ae5813e91c4e1e97570f8e41cf4'}

    start = datetime.now()
    mine_manager = MineManager(config_path=args.config)

    methods = mine_manager.file_miner.get_current_methods(file_query['hash'])

    m_changes = []
    for m in methods:
        changes = mine_manager.method_miner.get_change_history(m)
        mc = [{'complexity': x['complexity'],
               'date':  datetime.fromtimestamp(x['timestamp']),
               'name': m['name']} for x in changes]
        m_changes = m_changes + mc
    print('All methods complexity took: {}'.format(datetime.now() - start))
    print('Total methods: ', len(methods))

    if args.plot:
        df = pd.DataFrame(m_changes)
        df['date'] = pd.to_datetime(df.date)
        df = df.sort_values(by='date')
        fig = px.line(df, x="date", y="complexity", color="name",
                      line_group="name", hover_name="name")
        fig.show()


if __name__ == '__main__':
    main()



"""This module saves the cache data on disk"""
import collections
from diskcache import Index


class DrillCache:
    """Class for storing all data at once in the cache"""

    def __init__(self, data):
        """Transforms dictionary to ordered dic and saves it"""
        dt_ = [(k, v) for k, v in data.items()]
        self.data = Index(collections.OrderedDict(dt_))


class DrillCacheSequential:
    """Class for disk cache sequential"""

    def __init__(self):
        """Init drill cache"""
        self.data = Index([('commits', []),
                           ('parents', []), ('developers', []),
                           ('dev_commits', []), ('branches', []),
                           ('branches_commits', []), ('files', []),
                           ('commit_files', []), ('methods', []),
                           ('file_methods', []), ('commit_methods', [])
                           ])

    def append_cache(self, key, value):
        """Appends record to array on disk ccache
        :param key: data key
        :param value: value to append
        """
        temp_ = self.data[key]
        temp_.append(value)
        self.data[key] = temp_



"""Utils methods for miners"""


def format_commit_id_date(project_id, start_date, end_date, commit_hash=None):
    """Formats commit query with id and dates
    :param project_id: the project unique identifier
    :param start_date: timestamp, commit start_date
    :param end_date: timestamp, commit end_date
    :param ccommit_hash: optional, if given the query
      filters by commit hash
    :returns: query filter string and where clause
    """
    com_filter, where = "", ""
    if project_id and not commit_hash:
        com_filter += """{{project_id: "{0}"}}""".format(project_id)
    if project_id and commit_hash:
        com_filter += """{{project_id: "{0}", hash: {1}}}""".format(
            project_id, commit_hash)
    if start_date:
        where += "c.timestamp >= {0}".format(start_date)
    if end_date:
        where += " AND " if where else ""
        where += "c.timestamp <= {0}".format(end_date)
    where = "WHERE " + where if where else where

    return com_filter, where



# v0.2.8 released
from setuptools import setup, find_packages

with open('requirements.txt') as reqs_file:
    requirements = reqs_file.read().splitlines()

setup(name="graphrepo",
      version="0.2.8",
      description="A tool that maps a Github repo to Neo4j and Helps Mining the Repo in the DB",
      url="https://github.com/NullConvergence/GraphRepo",
      license='Apache License',
      python_requires='>=3.5',
      install_requires=requirements,
      packages=find_packages('.'),
      package_dir={'graphrepo': 'graphrepo'})

# python3 setup.py sdist bdist_wheel
# python3 -m twine upload dist/*



# Copyright 2019 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module is an example of indexing all data from a repository in Neo4j"""

import argparse
from graphrepo.drillers import Driller


def parse_args():
    """Parse argument"""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--config', default='examples/configs/pydriller.yml', type=str)
    return parser.parse_args()


def main():
    """Main method"""
    args = parse_args()
    driller = Driller(config_path=args.config)
    # this method should be called only once, when initializing
    # a database for the first time
    driller.init_db()
    driller.drill_batch()


if __name__ == '__main__':
    main()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pytest
import yaml

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.drillers.driller import Driller
from graphrepo.drillers.cache_driller import CacheDriller


class TestCommit:
    """Most data is indexed when indexing a commmit
    so this class tests indexing for multiple models"""

    def test_nodes_index(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        # test that all nodes were indexed
        node_matcher = NodeMatcher(test_driller.graph)
        all_commits = list(node_matcher.match("Commit"))
        assert len(all_commits) == 8

        all_devs = list(node_matcher.match("Developer"))
        assert len(all_devs) == 2

        all_files = list(node_matcher.match("File"))
        assert len(all_files) == 6

        all_methods = list(node_matcher.match("Method"))
        assert len(all_methods) == 5

        all_branches = list(node_matcher.match("Branch"))
        assert len(all_branches) == 1

        test_driller.clean()

    def test_rel_index(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        # test that all relationships were indexed
        rel_matcher = RelationshipMatcher(test_driller.graph)

        all_branch = list(rel_matcher.match(None, "BranchCommit"))
        assert len(all_branch) == 8

        all_authorship = list(rel_matcher.match(None, "Author"))
        assert len(all_authorship) == 8

        all_parent = list(rel_matcher.match(None, "Parent"))
        assert len(all_parent) == 8

        all_updadedfile = list(rel_matcher.match(None, "UpdateFile"))
        assert len(all_updadedfile) == 9

        all_hasmethod = list(rel_matcher.match(None, "Method"))
        assert len(all_hasmethod) == 5

        all_updatemethod = list(rel_matcher.match(None, "UpdateMethod"))
        assert len(all_updatemethod) == 9

        test_driller.clean()

    def test_rel_index_cache(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = CacheDriller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch_cache_sequential()

        # test that all relationships were indexed
        rel_matcher = RelationshipMatcher(test_driller.graph)

        all_branch = list(rel_matcher.match(None, "BranchCommit"))
        assert len(all_branch) == 8

        all_authorship = list(rel_matcher.match(None, "Author"))
        assert len(all_authorship) == 8

        all_parent = list(rel_matcher.match(None, "Parent"))
        assert len(all_parent) == 8

        all_updadedfile = list(rel_matcher.match(None, "UpdateFile"))
        assert len(all_updadedfile) == 9

        all_hasmethod = list(rel_matcher.match(None, "Method"))
        assert len(all_hasmethod) == 5

        all_updatemethod = list(rel_matcher.match(None, "UpdateMethod"))
        assert len(all_updatemethod) == 9

        test_driller.clean()

    def test_custom_attributes_rel(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        node_matcher = NodeMatcher(test_driller.graph)
        rel_matcher = RelationshipMatcher(test_driller.graph)

        commit = node_matcher.match(
            "Commit", hash="aa6fa504ccb0fa919acc3cb31e510dc2048314eb0656f34babada15c").first()
        assert commit['is_merge'] == 0

        update_file_rel = rel_matcher.match([commit], "UpdateFile").first()
        assert update_file_rel['complexity'] == 2
        assert update_file_rel['nloc'] == 8
        assert update_file_rel['old_path'] == 'gr_test/default_class.py'
        assert update_file_rel['path'] == 'gr_test/default_class.py'
        assert update_file_rel['token_count'] == 42
        assert update_file_rel['type'] == 'MODIFY'
        assert update_file_rel['removed'] == 6
        assert update_file_rel['added'] == 0

        update_method_rel = rel_matcher.match(
            [commit], 'UpdateMethod').first()
        # assert update_method_rel['type'] == 'DELETE'
        assert update_method_rel['nloc'] == 5
        assert update_method_rel['complexity'] == 2
        assert update_method_rel['token_count'] == 21
        assert update_method_rel['length'] == 5
        assert update_method_rel['fan_in'] == 0
        assert update_method_rel['fan_out'] == 0
        assert update_method_rel['start_line'] == 11
        assert update_method_rel['end_line'] == 15

        test_driller.clean()



# Copyright 2020 NullConvergence
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
import os

from py2neo import NodeMatcher, RelationshipMatcher
from graphrepo.drillers.driller import Driller
from graphrepo.miners.commit import CommitMiner


class TestCommitMiner:
    def test_gets(self):
        folder = os.path.dirname(os.path.abspath(__file__))
        test_driller = Driller(os.path.join(folder, 'cnfg_simple.yml'))
        test_driller.drill_batch()

        st_date = datetime.strptime(
            '14 May, 2020 00:00', '%d %B, %Y %H:%M').timestamp()
        end_date = datetime.strptime(
            '15 May, 2020 02:00', '%d %B, %Y %H:%M').timestamp()

        n_matcher = NodeMatcher(test_driller.graph)
        r_matcher = RelationshipMatcher(test_driller.graph)

        com_miner = CommitMiner(test_driller.graph, n_matcher, r_matcher)

        all_com = com_miner.get_all()
        assert len(all_com) == 8

        all_com_dates = com_miner.get_between_dates(st_date, end_date)
        assert len(all_com_dates) == 8

        c_files = com_miner.get_commit_files(
            'ad98f8594c15b1ebc4be4f20d849bcc0edf69ec574c33dfd84b7792d')
        assert len(c_files) == 3

        c_file_updates = com_miner.get_commit_file_updates(
            'ad98f8594c15b1ebc4be4f20d849bcc0edf69ec574c33dfd84b7792d')
        assert len(c_file_updates) == 3

        c_methods = com_miner.get_commit_methods(
            'ad98f8594c15b1ebc4be4f20d849bcc0edf69ec574c33dfd84b7792d')
        assert len(c_methods) == 3

        c_method_updates = com_miner.get_commit_method_updates(
            'ad98f8594c15b1ebc4be4f20d849bcc0edf69ec574c33dfd84b7792d')
        assert len(c_method_updates) == 3

        test_driller.clean()
